import random
import typing

from tests.container import DIContainer
from that_depends import BaseContainer, providers


def _sync_resource() -> typing.Iterator[float]:
    yield random.random()


async def test_container_sync_teardown() -> None:
    await DIContainer.init_resources()
    DIContainer.tear_down_sync()
    for provider in DIContainer.providers.values():
        if isinstance(provider, providers.Resource):
            if provider._is_async:
                assert provider._context.instance is not None
            else:
                assert provider._context.instance is None
        if isinstance(provider, providers.Singleton):
            assert provider._instance is None


async def test_container_tear_down() -> None:
    await DIContainer.init_resources()
    await DIContainer.tear_down()
    for provider in DIContainer.providers.values():
        if isinstance(provider, providers.Resource):
            assert provider._context.instance is None
        if isinstance(provider, providers.Singleton):
            assert provider._instance is None


async def test_container_sync_tear_down_propagation() -> None:
    class _DependentContainer(BaseContainer):
        singleton = providers.Singleton(lambda: random.random())
        resource = providers.Resource(_sync_resource)

    DIContainer.connect_containers(_DependentContainer)

    await DIContainer.init_resources()

    assert isinstance(_DependentContainer.singleton._instance, float)
    assert isinstance(_DependentContainer.resource._context.instance, float)

    DIContainer.tear_down_sync()

    assert _DependentContainer.singleton._instance is None
    assert _DependentContainer.resource._context.instance is None


async def test_container_tear_down_propagation() -> None:
    async def _async_singleton() -> float:
        return random.random()

    async def _async_resource() -> typing.AsyncIterator[float]:
        yield random.random()

    class _DependentContainer(BaseContainer):
        async_singleton = providers.AsyncSingleton(_async_singleton)
        async_resource = providers.Resource(_async_resource)
        sync_singleton = providers.Singleton(lambda: random.random())
        sync_resource = providers.Resource(_sync_resource)

    DIContainer.connect_containers(_DependentContainer)

    await DIContainer.init_resources()

    assert isinstance(_DependentContainer.async_singleton._instance, float)
    assert isinstance(_DependentContainer.async_resource._context.instance, float)
    assert isinstance(_DependentContainer.sync_singleton._instance, float)
    assert isinstance(_DependentContainer.sync_resource._context.instance, float)

    await DIContainer.tear_down()

    assert _DependentContainer.async_singleton._instance is None
    assert _DependentContainer.async_resource._context.instance is None
    assert _DependentContainer.sync_singleton._instance is None
    assert _DependentContainer.sync_resource._context.instance is None



"""Providers."""

from that_depends.providers.base import AbstractProvider, AttrGetter
from that_depends.providers.collection import Dict, List
from that_depends.providers.context_resources import (
    ContextResource,
    DIContextMiddleware,
    container_context,
)
from that_depends.providers.factories import AsyncFactory, Factory
from that_depends.providers.local_singleton import ThreadLocalSingleton
from that_depends.providers.object import Object
from that_depends.providers.resources import Resource
from that_depends.providers.selector import Selector
from that_depends.providers.singleton import AsyncSingleton, Singleton


__all__ = [
    "AbstractProvider",
    "AsyncFactory",
    "AsyncSingleton",
    "AttrGetter",
    "ContextResource",
    "DIContextMiddleware",
    "Dict",
    "Factory",
    "List",
    "Object",
    "Resource",
    "Selector",
    "Singleton",
    "ThreadLocalSingleton",
    "container_context",
]



import asyncio
import threading
import typing

from typing_extensions import override

from that_depends.providers import AbstractProvider
from that_depends.providers.mixin import SupportsTeardown


T_co = typing.TypeVar("T_co", covariant=True)
P = typing.ParamSpec("P")


class ThreadLocalSingleton(SupportsTeardown, AbstractProvider[T_co]):
    """Creates a new instance for each thread using a thread-local store.

    This provider ensures that each thread gets its own instance, which is
    created via the specified factory function. Once created, the instance is
    cached for future injections within the same thread.

    Example:
        ```python
        def factory():
            return random.randint(1, 100)

        singleton = ThreadLocalSingleton(factory)

        # Same thread, same instance
        instance1 = singleton.sync_resolve()
        instance2 = singleton.sync_resolve()

        def thread_task():
            return singleton.sync_resolve()

        threads = [threading.Thread(target=thread_task) for i in range(10)]
        for thread in threads:
            thread.start() # Each thread will get a different instance
        ```

    """

    def __init__(self, factory: typing.Callable[P, T_co], *args: P.args, **kwargs: P.kwargs) -> None:
        """Initialize the ThreadLocalSingleton provider.

        Args:
            factory: A callable that returns a new instance of the dependency.
            *args: Positional arguments to pass to the factory.
            **kwargs: Keyword arguments to pass to the factory

        """
        super().__init__()
        self._factory: typing.Final = factory
        self._thread_local = threading.local()
        self._asyncio_lock = asyncio.Lock()
        self._args: typing.Final = args
        self._kwargs: typing.Final = kwargs

    def _register_arguments(self) -> None:
        self._register(self._args)
        self._register(self._kwargs.values())

    def _deregister_arguments(self) -> None:
        self._deregister(self._args)
        self._deregister(self._kwargs.values())

    @property
    def _instance(self) -> T_co | None:
        return getattr(self._thread_local, "instance", None)

    @_instance.setter
    def _instance(self, value: T_co | None) -> None:
        self._thread_local.instance = value

    @override
    async def resolve(self) -> T_co:
        if self._override is not None:
            return typing.cast(T_co, self._override)

        async with self._asyncio_lock:
            if self._instance is not None:
                return self._instance

            self._register_arguments()

            self._instance = self._factory(
                *[await x.resolve() if isinstance(x, AbstractProvider) else x for x in self._args],  # type: ignore[arg-type]
                **{  # type: ignore[arg-type]
                    k: await v.resolve() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()
                },
            )
            return self._instance

    @override
    def resolve_sync(self) -> T_co:
        if self._override is not None:
            return typing.cast(T_co, self._override)

        if self._instance is not None:
            return self._instance

        self._register_arguments()

        self._instance = self._factory(
            *[x.resolve_sync() if isinstance(x, AbstractProvider) else x for x in self._args],  # type: ignore[arg-type]
            **{k: v.resolve_sync() if isinstance(v, AbstractProvider) else v for k, v in self._kwargs.items()},  # type: ignore[arg-type]
        )
        return self._instance

    @override
    def tear_down_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        if self._instance is not None:
            self._instance = None
        self._deregister_arguments()
        if propagate:
            self._tear_down_children_sync(propagate=propagate, raise_on_async=raise_on_async)

    @override
    async def tear_down(self, propagate: bool = True) -> None:
        """Reset the thread-local instance.

        After calling this method, subsequent calls to `sync_resolve` on the
        same thread will produce a new instance.
        """
        if self._instance is not None:
            self._instance = None
        self._deregister_arguments()
        if propagate:
            await self._tear_down_children()



import datetime

import pytest

from tests import container
from that_depends import BaseContainer, providers


class DIContainer(BaseContainer):
    alias = "dynamic_container"
    sync_resource: providers.Resource[datetime.datetime]


async def test_dynamic_container_not_supported() -> None:
    new_provider = providers.Resource(container.create_sync_resource)
    with pytest.raises(AttributeError):
        DIContainer.sync_resource = new_provider

    with pytest.raises(AttributeError):
        DIContainer.something_new = new_provider



import typing

import pytest

from tests.container import create_async_resource, create_sync_resource
from that_depends import BaseContainer, providers
from that_depends.providers import AbstractProvider


class DIContainer(BaseContainer):
    alias = "collection_container"
    sync_resource = providers.Resource(create_sync_resource)
    async_resource = providers.Resource(create_async_resource)
    sequence = providers.List(sync_resource, async_resource)
    mapping = providers.Dict(sync_resource=sync_resource, async_resource=async_resource)


@pytest.fixture(autouse=True)
async def _clear_di_container() -> typing.AsyncIterator[None]:
    try:
        yield
    finally:
        await DIContainer.tear_down()


async def test_list_provider() -> None:
    sequence = await DIContainer.sequence()
    sync_resource = await DIContainer.sync_resource()
    async_resource = await DIContainer.async_resource()

    assert sequence == [sync_resource, async_resource]


def test_list_failed_sync_resolve() -> None:
    with pytest.raises(RuntimeError, match="AsyncResource cannot be resolved synchronously"):
        DIContainer.sequence.resolve_sync()


async def test_list_sync_resolve_after_init() -> None:
    await DIContainer.init_resources()
    DIContainer.sequence.resolve_sync()


async def test_dict_provider() -> None:
    mapping = await DIContainer.mapping()
    sync_resource = await DIContainer.sync_resource()
    async_resource = await DIContainer.async_resource()

    assert mapping == {"sync_resource": sync_resource, "async_resource": async_resource}
    assert mapping == DIContainer.mapping.resolve_sync()


@pytest.mark.parametrize("provider", [DIContainer.sequence, DIContainer.mapping])
async def test_attr_getter_in_collections_providers(provider: AbstractProvider[typing.Any]) -> None:
    with pytest.raises(AttributeError, match=f"'{type(provider)}' object has no attribute 'some_attribute'"):
        await provider.some_attribute



import asyncio
import random
import typing

import pytest
from typing_extensions import override

from that_depends import BaseContainer
from that_depends.providers.base import AbstractProvider
from that_depends.providers.local_singleton import ThreadLocalSingleton
from that_depends.providers.mixin import SupportsTeardown
from that_depends.providers.resources import Resource
from that_depends.providers.singleton import AsyncSingleton, Singleton


class DummyProvider(SupportsTeardown, AbstractProvider[int]):
    """A dummy provider used for testing."""

    def __init__(self) -> None:
        super().__init__()
        self._instance: int | None = None

    @override
    async def tear_down(self, propagate: bool = True) -> None:
        self._instance = None
        if propagate:
            await self._tear_down_children()

    @override
    def tear_down_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        self._instance = None
        if propagate:
            self._tear_down_children_sync(propagate=propagate, raise_on_async=raise_on_async)

    @override
    async def resolve(self) -> int:
        self._instance = 1
        return self._instance

    @override
    def resolve_sync(self) -> int:
        self._instance = 1
        return self._instance  # pragma: no cover


def test_add_child_provider() -> None:
    provider_a = DummyProvider()
    provider_b = DummyProvider()

    assert len(provider_a._children) == 0, "Expected provider_a to have no children initially"
    provider_a.add_child_provider(provider_b)

    assert len(provider_a._children) == 1, "Expected provider_a._children to have 1 child"
    assert provider_b in provider_a._children, "provider_b should be in provider_a._children"


def test_register_with_providers() -> None:
    parent = DummyProvider()
    child_1 = DummyProvider()
    child_2 = DummyProvider()

    parent._register([child_1, child_2])

    assert parent in child_1._children, "Expected child_1._children to contain parent"
    assert parent in child_2._children, "Expected child_2._children to contain parent"


def test_register_with_mixed_items() -> None:
    parent = DummyProvider()
    child_1 = DummyProvider()
    non_provider = 12345  # not an AbstractProvider

    parent._register([child_1, non_provider])

    assert parent in child_1._children, "Expected child_1._children to contain parent"


def test_sync_tear_down_propagation() -> None:
    parent = DummyProvider()
    child_1 = DummyProvider()
    child_2 = DummyProvider()

    parent.add_child_provider(child_1)
    parent.add_child_provider(child_2)

    assert parent.resolve_sync() == 1

    assert parent._instance == 1

    parent.tear_down_sync()

    assert parent._instance is None
    assert child_1._instance is None
    assert child_2._instance is None


async def test_async_tear_down_propagation() -> None:
    parent = DummyProvider()
    child_1 = DummyProvider()
    child_2 = DummyProvider()

    parent.add_child_provider(child_1)
    parent.add_child_provider(child_2)

    assert await parent.resolve() == 1

    assert parent._instance == 1

    await parent.tear_down()

    assert parent._instance is None
    assert child_1._instance is None
    assert child_2._instance is None


_RETURN_VALUE = 42


def _simple_factory_value() -> int:
    return _RETURN_VALUE


async def _simple_async_factory_value(v: int) -> int:
    await asyncio.sleep(0.001)
    return v


def _resource_generator(v: int) -> typing.Iterator[int]:
    """Sync generator resource with teardown logic."""
    try:
        yield v
    finally:
        # Normally you'd close files, DB connections, etc. here
        pass


@pytest.fixture
def dummy_singleton() -> Singleton[int]:
    """Get sync factory for test use."""
    return Singleton(_simple_factory_value)


def test_singleton_registration_and_deregistration(dummy_singleton: Singleton[int]) -> None:
    singleton = Singleton(lambda x: x + 1, dummy_singleton.cast)
    assert singleton not in dummy_singleton._children, "Singleton should not be registered as child yet."
    singleton.resolve_sync()

    assert singleton in dummy_singleton._children, "Singleton should be registered as a child."

    dummy_singleton.tear_down_sync()

    assert singleton not in dummy_singleton._children, (
        "Singleton should be removed from parent's children after tear_down."
    )


def test_thread_local_singleton_registration_and_deregistration(dummy_singleton: Singleton[int]) -> None:
    thread_local = ThreadLocalSingleton(lambda val: f"TL-{val}", dummy_singleton)

    assert thread_local not in dummy_singleton._children, "ThreadLocalSingleton not registered as child."

    thread_local.resolve_sync()
    assert thread_local in dummy_singleton._children, "ThreadLocalSingleton not registered as child."

    # Teardown
    thread_local.tear_down_sync()

    assert thread_local not in dummy_singleton._children, "ThreadLocalSingleton should be deregistered after teardown."


def test_resource_registration_and_deregistration(dummy_singleton: Singleton[int]) -> None:
    resource = Resource(_resource_generator, dummy_singleton.cast)

    assert resource not in dummy_singleton._children, "Resource should not be registered as child yet."

    resource.resolve_sync()

    assert resource in dummy_singleton._children, "Resource should be registered as child."

    resource.tear_down_sync()
    assert resource not in dummy_singleton._children, "Resource should be deregistered after teardown."


async def test_async_singleton_registration_and_deregistration(dummy_singleton: Singleton[int]) -> None:
    async_singleton = AsyncSingleton(_simple_async_factory_value, dummy_singleton.cast)

    await async_singleton.resolve()

    assert async_singleton in dummy_singleton._children

    value = await async_singleton.resolve()
    assert value == _RETURN_VALUE

    await async_singleton.tear_down()

    assert async_singleton not in dummy_singleton._children


def test_teardown_propagation_chain() -> None:
    def _sync_resource_gen(v: float) -> typing.Iterator[float]:
        try:
            yield v
        finally:
            pass

    def _grandchild_gen() -> typing.Iterator[float]:
        try:
            yield random.random()
        finally:
            pass

    parent_resource = Resource(_grandchild_gen)
    child_singleton = Singleton(lambda g: g + 1.0, parent_resource.cast)
    grandchild = Resource(_sync_resource_gen, child_singleton.cast)

    parent_value = parent_resource.resolve_sync()
    grandchild_value = grandchild.resolve_sync()

    assert child_singleton in parent_resource._children
    assert grandchild in child_singleton._children
    parent_resource.tear_down_sync(propagate=True)

    assert grandchild.resolve_sync() != grandchild_value
    assert parent_resource.resolve_sync() != parent_value


def test_propagate_off() -> None:
    parent = Singleton(_simple_factory_value)
    child = Singleton(lambda x: x + 1, parent.cast)

    child.resolve_sync()

    parent.tear_down_sync(propagate=False)

    assert child in parent._children
    assert child._instance is not None


async def test_async_tear_down_propagation_with_singleton() -> None:
    parent = Singleton(_simple_factory_value)
    child = Singleton(lambda x: x + 1, parent.cast)

    await child.resolve()

    await parent.tear_down()

    assert child._instance is None


async def test_async_propagate_off() -> None:
    parent = Singleton(_simple_factory_value)
    child = Singleton(lambda x: x + 1, parent.cast)

    await child.resolve()

    await parent.tear_down(propagate=False)

    assert child._instance is not None


async def test_provider_registration_in_different_scope_async() -> None:
    async def _creator() -> int:
        return 1

    async def _identity(x: int) -> int:
        return x

    class Container(BaseContainer):
        provider = AsyncSingleton(_creator)

    async def nested() -> int:
        p = AsyncSingleton(_identity, Container.provider.cast)

        result = await p.resolve()

        await p.tear_down()
        assert len(p._parents) == 0
        return result

    await nested()

    assert len(Container.provider._children) == 0


def test_provider_registration_in_different_scope_sync() -> None:
    class Container(BaseContainer):
        provider = Singleton(lambda: 1)

    def nested() -> int:
        p = Singleton(lambda x: x, Container.provider.cast)

        result = p.resolve_sync()

        p.tear_down_sync()
        assert len(p._parents) == 0
        return result

    nested()

    assert len(Container.provider._children) == 0



import asyncio
import contextlib
import threading
import typing
import warnings

from typing_extensions import override

from that_depends.providers.mixin import CannotTearDownSyncError, SupportsTeardown


T_co = typing.TypeVar("T_co", covariant=True)


class ResourceContext(SupportsTeardown, typing.Generic[T_co]):
    """Class to manage a resources' context."""

    __slots__ = "asyncio_lock", "context_stack", "instance", "is_async", "threading_lock"

    def __init__(self, is_async: bool) -> None:
        """Create a new ResourceContext instance.

        Args:
            is_async (bool): Whether the ResourceContext was created in
                an async context.
        For example within a ``async with container_context(Container): ...`` statement.

        """
        self.instance: T_co | None = None
        self.asyncio_lock: typing.Final = asyncio.Lock()
        self.threading_lock: typing.Final = threading.Lock()
        self.context_stack: contextlib.AsyncExitStack | contextlib.ExitStack | None = None
        self.is_async = is_async

    @staticmethod
    def is_context_stack_async(
        context_stack: contextlib.AsyncExitStack | contextlib.ExitStack | None,
    ) -> typing.TypeGuard[contextlib.AsyncExitStack]:
        """Check if the context stack is an async context stack."""
        return isinstance(context_stack, contextlib.AsyncExitStack)

    @staticmethod
    def is_context_stack_sync(
        context_stack: contextlib.AsyncExitStack | contextlib.ExitStack,
    ) -> typing.TypeGuard[contextlib.ExitStack]:
        """Check if the context stack is a sync context stack."""
        return isinstance(context_stack, contextlib.ExitStack)

    @override
    async def tear_down(self, propagate: bool = True) -> None:
        """Tear down the async context stack."""
        if self.context_stack is None:
            return

        if self.is_context_stack_async(self.context_stack):
            await self.context_stack.aclose()
        elif self.is_context_stack_sync(self.context_stack):
            self.context_stack.close()
        self.context_stack = None
        self.instance = None

    @override
    def tear_down_sync(self, propagate: bool = True, raise_on_async: bool = True) -> None:
        """Tear down the sync context stack."""
        if self.context_stack is None:
            return

        if self.is_context_stack_sync(self.context_stack):
            self.context_stack.close()
            self.context_stack = None
            self.instance = None
        elif self.is_context_stack_async(self.context_stack):
            msg = "Cannot tear down async context in sync mode"
            if raise_on_async:
                raise CannotTearDownSyncError(msg)
            warnings.warn(msg, RuntimeWarning, stacklevel=2)
